// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Main program, simply calls D_DoomMain high level loop.
//
//-----------------------------------------------------------------------------

static const char
rcsid[] = "$Id: i_main.c,v 1.4 1997/02/03 22:45:10 b1 Exp $";

#define _WIN32_WINNT 0x0500     // for Mouse Wheei support (@uvO~OWindows()vP358)
#include <windows.h>
#include "doomdef.h"
#include "m_argv.h"
#include "d_main.h"
#include "i_system.h"
#include "music.h"
#include "m_misc.h"

/*
int
main
( int		argc,
  char**	argv ) 
{ 
    myargc = argc; 
    myargv = argv; 
 
    D_DoomMain (); 

    return 0;
} 
*/

HDC     hDC;
HWND    hMainWnd;
int     screenwidth;
int     screenheight;
int     replay;

extern HGLRC hRC;

void D_DoomLoop(void);

int     buttonstate;
int     mousewheel;

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	event_t event;

	switch(message)
	{
	case WM_KEYDOWN:
		if (lParam & 0x40000000)
			return 0;
		event.type = ev_keydown;
		event.data1 = wParam/*tolower(wParam)*/; // @̃At@xbgascii codeFL[Virtual-Key Code_u
		D_PostEvent(&event);
		return 0;

	case WM_KEYUP:
		event.type = ev_keyup;
		event.data1 = wParam/*tolower(wParam)*/;
		D_PostEvent(&event);
		return 0;
	
    case WM_SYSKEYDOWN: // @AltL[F10L[ꂽBMSDNCu "WM_SYSKEYDOWN" Q
        if (lParam & 0x40000000)
			return 0;
        if (wParam == VK_F10)
        {
			event.type = ev_keydown;
            event.data1 = wParam;
            D_PostEvent(&event);
            return 0;
        }
		break;

    // @QUAKE\[X(MainWndProc()[gl_vidnt.c])甲
    // this is complicated because Win32 seems to pack multiple mouse events into
	// one update sometimes, so we always check all states and look for events
	case WM_LBUTTONDOWN:
	case WM_LBUTTONUP:
	case WM_RBUTTONDOWN:
	case WM_RBUTTONUP:
	case WM_MBUTTONDOWN:
	case WM_MBUTTONUP:
	case WM_MOUSEMOVE:
		buttonstate = 0;
		if (wParam & MK_LBUTTON)
			buttonstate |= 1;
		if (wParam & MK_RBUTTON)
			buttonstate |= 2;
		if (wParam & MK_MBUTTON)
			buttonstate |= 4;
		event.type = ev_mouse;
        event.data1 = buttonstate;
        event.data2 = event.data3 = 0;
        D_PostEvent(&event);
		return 0;

    case WM_MOUSEWHEEL: 
        if (!mousewheel)
            break;
		if ((short)HIWORD(wParam) > 0) 
            event.data3 = 50;   // MAXPLMOVE = 0x32
        else 
            event.data3 = -50;
        event.type = ev_mouse;
        event.data1 = buttonstate;
        event.data2 = 0;
        D_PostEvent(&event);
		return 0;

    case MM_MCINOTIFY:
        if (wParam == MCI_NOTIFY_SUCCESSFUL)
        {
            if (replay)
                ReplayMidi();
            else
                StopMidi();
        }
        return 0;

	case WM_DESTROY:
		I_Quit();
		//PostQuitMessage(0);
		return 0;
	}

	return DefWindowProc(hwnd, message, wParam, lParam);
}

extern int  debugout;
extern int	key_right;
extern int	key_left;
extern int	key_up;
extern int	key_down;
extern int	key_strafeleft;
extern int	key_straferight;
extern int  mousewalk;
extern int  midbuttonuse;
extern int  rebuildbsp;
extern int  nowait;

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow)
{
	HWND        hwnd;
	MSG         msg;
	WNDCLASS    wndclass;
	char        szAppName[] = "GLDoom";
    FILE*       fp;
    int         p;
    int         cxscreen, cyscreen;
    BOOL        dispchange = FALSE;
	
    ParseCommand(szCmdLine); // @szCmdLine̓vO͊܂܂Ȃ

    M_LoadDefaults();

    p = M_CheckParm("-debugout");
    if (p && p < myargc-1)
        debugout = atoi(myargv[p+1]);
    if (debugout)
    {
        fp = fopen(debugname, "w");
        fclose(fp);
    }

    p = M_CheckParm("-turnright");
    if (p && p < myargc-1)
        key_right = myargv[p+1][0];
    p = M_CheckParm("-turnleft");
    if (p && p < myargc-1)
        key_left = myargv[p+1][0];
    p = M_CheckParm("-walkforth");
    if (p && p < myargc-1)
        key_up = myargv[p+1][0];
    p = M_CheckParm("-walkback");
    if (p && p < myargc-1)
        key_down = myargv[p+1][0];
    p = M_CheckParm("-strafeleft");
    if (p && p < myargc-1)
        key_strafeleft = myargv[p+1][0];
    p = M_CheckParm("-straferight");
    if (p && p < myargc-1)
        key_straferight = myargv[p+1][0];
    p = M_CheckParm("-mousewheel");
    if (p && p < myargc-1)
        mousewheel = atoi(myargv[p+1]);
    p = M_CheckParm("-mousewalk");
    if (p && p < myargc-1)
        mousewalk = atoi(myargv[p+1]);
    p = M_CheckParm("-midbuttonuse");
    if (p && p < myargc-1)
        midbuttonuse = atoi(myargv[p+1]);
    p = M_CheckParm("-rebuild");
    if (p && p < myargc-1)
        rebuildbsp = atoi(myargv[p+1]);
    p = M_CheckParm("-nowait");
    if (p && p < myargc-1)
        nowait = atoi(myargv[p+1]);
    p = M_CheckParm("-width");
    if (p && p < myargc-1)
        screenwidth = atoi(myargv[p+1]);
    p = M_CheckParm("-height");
    if (p && p < myargc-1)
        screenheight = atoi(myargv[p+1]);

    cxscreen = GetSystemMetrics(SM_CXSCREEN);
    cyscreen = GetSystemMetrics(SM_CYSCREEN);

    if (!screenwidth || !screenheight)
    {
        screenwidth = cxscreen;
        screenheight = cyscreen;
    }

    if (screenwidth != cxscreen || screenheight != cyscreen)
    {
        DEVMODE devmode;
        LONG ret;

	    devmode.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT;
	    devmode.dmPelsWidth = screenwidth;
	    devmode.dmPelsHeight = screenheight;
	    devmode.dmSize = sizeof(devmode);

	    ret = ChangeDisplaySettings(&devmode, CDS_FULLSCREEN); // @Platform SDK Q
        if (ret != DISP_CHANGE_SUCCESSFUL)
            I_Error("ChangeDisplaySettings() failed");

        dispchange = TRUE;
    }

	wndclass.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wndclass.lpfnWndProc = WndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(hInstance, "sposa1");
	wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = szAppName;

	RegisterClass(&wndclass);

	hwnd = CreateWindow(szAppName,
	                    szAppName,
	                    WS_POPUP | WS_VISIBLE,
	                    0,
	                    0,
	                    screenwidth,
	                    screenheight,
	                    NULL,
	                    NULL,
	                    hInstance,
	                    NULL);

	ShowCursor(FALSE);
	hDC = GetDC(hwnd);
	hMainWnd = hwnd;
	D_DoomMain ();

	while (1)
	{
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if (msg.message == WM_QUIT)
				break;

			DispatchMessage(&msg);
		} else
			D_DoomLoop();
	}

    if (dispchange)
        ChangeDisplaySettings(NULL, 0);

	return msg.wParam;
}
